@@ -18,6 +18,7 @@ import fixtures
 import mock
 from oslotest import base
 from six import moves
+import tempfile
 import testscenarios
 
 from oslo_config import cfg
@@ -729,9 +730,9 @@ class GeneratorTestCase(base.BaseTestCase):
     def _capture_stdout(self):
         return self._capture_stream('stdout')
 
-    @mock.patch('stevedore.named.NamedExtensionManager')
+    @mock.patch.object(generator, '_get_raw_opts_loaders')
     @mock.patch.object(generator, 'LOG')
-    def test_generate(self, mock_log, named_mgr):
+    def test_generate(self, mock_log, raw_opts_loader):
         generator.register_cli_opts(self.conf)
 
         namespaces = [i[0] for i in self.opts]
@@ -750,12 +751,17 @@ class GeneratorTestCase(base.BaseTestCase):
             output_file = self.tempdir.join(self.output_file)
             self.config(output_file=output_file)
 
-        mock_eps = []
-        for name, opts in self.opts:
-            mock_ep = mock.Mock()
-            mock_ep.configure_mock(name=name, obj=opts)
-            mock_eps.append(mock_ep)
-        named_mgr.return_value = mock_eps
+        # We have a static data structure matching what should be
+        # returned by _list_opts() but we're mocking out a lower level
+        # function that needs to return a namespace and a callable to
+        # return options from that namespace. We have to pass opts to
+        # the lambda to cache a reference to the name because the list
+        # comprehension changes the thing pointed to by the name each
+        # time through the loop.
+        raw_opts_loader.return_value = [
+            (ns, lambda opts=opts: opts)
+            for ns, opts in self.opts
+        ]
 
         generator.generate(self.conf)
 
@@ -765,12 +771,6 @@ class GeneratorTestCase(base.BaseTestCase):
             content = open(output_file).read()
             self.assertEqual(self.expected, content)
 
-        named_mgr.assert_called_once_with(
-            'oslo.config.opts',
-            names=namespaces,
-            on_load_failure_callback=generator.on_load_failure_callback,
-            invoke_on_load=True)
-
         log_warning = getattr(self, 'log_warning', None)
         if log_warning is not None:
             mock_log.warning.assert_called_once_with(*log_warning)
@@ -780,28 +780,150 @@ class GeneratorTestCase(base.BaseTestCase):
 
 class IgnoreDoublesTestCase(base.BaseTestCase):
 
-    @mock.patch('stevedore.named.NamedExtensionManager')
-    def test_list_ignores_doubles(self, named_mgr):
-        config_opts = [cfg.StrOpt('foo'),
-                       cfg.StrOpt('bar'),
-                       ]
-        mock_ep1 = mock.Mock()
-        mock_ep1.configure_mock(name="namespace",
-                                obj=[("group", config_opts)])
-        mock_ep2 = mock.Mock()
-        mock_ep2.configure_mock(name="namespace",
-                                obj=[("group", config_opts)])
+    opts = [cfg.StrOpt('foo', help='foo option'),
+            cfg.StrOpt('bar', help='bar option'),
+            cfg.StrOpt('foo_bar', help='foobar'),
+            cfg.StrOpt('str_opt', help='a string'),
+            cfg.BoolOpt('bool_opt', help='a boolean'),
+            cfg.IntOpt('int_opt', help='an integer')]
+
+    def test_cleanup_opts_default(self):
+        o = [("namespace1", [
+              ("group1", self.opts)])]
+        self.assertEqual(o, generator._cleanup_opts(o))
+
+    def test_cleanup_opts_dup_opt(self):
+        o = [("namespace1", [
+              ("group1", self.opts + [self.opts[0]])])]
+        e = [("namespace1", [
+              ("group1", self.opts)])]
+        self.assertEqual(e, generator._cleanup_opts(o))
+
+    def test_cleanup_opts_dup_groups_opt(self):
+        o = [("namespace1", [
+              ("group1", self.opts + [self.opts[1]]),
+              ("group2", self.opts),
+              ("group3", self.opts + [self.opts[2]])])]
+        e = [("namespace1", [
+              ("group1", self.opts),
+              ("group2", self.opts),
+              ("group3", self.opts)])]
+        self.assertEqual(e, generator._cleanup_opts(o))
+
+    def test_cleanup_opts_dup_namespace_groups_opts(self):
+        o = [("namespace1", [
+              ("group1", self.opts + [self.opts[1]]),
+              ("group2", self.opts)]),
+             ("namespace2", [
+              ("group1", self.opts + [self.opts[2]]),
+              ("group2", self.opts)])]
+        e = [("namespace1", [
+              ("group1", self.opts),
+              ("group2", self.opts)]),
+             ("namespace2", [
+              ("group1", self.opts),
+              ("group2", self.opts)])]
+        self.assertEqual(e, generator._cleanup_opts(o))
+
+    @mock.patch.object(generator, '_get_raw_opts_loaders')
+    def test_list_ignores_doubles(self, raw_opts_loaders):
+        config_opts = [
+            (None, [cfg.StrOpt('foo'), cfg.StrOpt('bar')]),
+        ]
+
         # These are the very same config options, but read twice.
         # This is possible if one misconfigures the entry point for the
         # sample config generator.
-        mock_eps = [mock_ep1, mock_ep2]
-        named_mgr.return_value = mock_eps
+        raw_opts_loaders.return_value = [
+            ('namespace', lambda: config_opts),
+            ('namespace', lambda: config_opts),
+        ]
 
         slurped_opts = 0
         for _, listing in generator._list_opts(None):
             for _, opts in listing:
                 slurped_opts += len(opts)
-        self.assertEqual(len(config_opts), slurped_opts)
+        self.assertEqual(2, slurped_opts)
+
+
+class GeneratorAdditionalTestCase(base.BaseTestCase):
+
+    opts = [cfg.StrOpt('foo', help='foo option', default='fred'),
+            cfg.StrOpt('bar', help='bar option'),
+            cfg.StrOpt('foo_bar', help='foobar'),
+            cfg.StrOpt('str_opt', help='a string'),
+            cfg.BoolOpt('bool_opt', help='a boolean'),
+            cfg.IntOpt('int_opt', help='an integer')]
+
+    def test_get_group_name(self):
+        name = "group1"
+        item = [name]
+        self.assertEqual(name, generator._get_group_name(item))
+
+    def test_get_group_name_as_optgroup(self):
+        name = "group2"
+        item = [cfg.OptGroup(name)]
+        self.assertEqual(name, generator._get_group_name(item))
+
+    def test_get_groups_empty_ns(self):
+        groups = generator._get_groups([])
+        self.assertEqual({'DEFAULT': []}, groups)
+
+    def test_get_groups_single_ns(self):
+        config = [("namespace1", [
+                   ("beta", self.opts),
+                   ("alpha", self.opts)])]
+        groups = generator._get_groups(config)
+        self.assertEqual(['DEFAULT', 'alpha', 'beta'], sorted(groups))
+
+    def test_get_groups_multiple_ns(self):
+        config = [("namespace1", [
+                   ("beta", self.opts),
+                   ("alpha", self.opts)]),
+                  ("namespace2", [
+                   ("gamma", self.opts),
+                   ("alpha", self.opts)])]
+        groups = generator._get_groups(config)
+        self.assertEqual(['DEFAULT', 'alpha', 'beta', 'gamma'], sorted(groups))
+
+    def test_output_opts_empty_default(self):
+
+        config = [("namespace1", [
+                   ("alpha", [])])]
+        groups = generator._get_groups(config)
+
+        fd, tmp_file = tempfile.mkstemp()
+        f = open(tmp_file, 'w+')
+        formatter = generator._OptFormatter(output_file=f)
+        expected = '''[DEFAULT]
+'''
+        generator._output_opts(formatter, 'DEFAULT', groups.pop('DEFAULT'))
+        f.close()
+        content = open(tmp_file).read()
+        self.assertEqual(expected, content)
+
+    def test_output_opts_group(self):
+
+        config = [("namespace1", [
+                   ("alpha", [self.opts[0]])])]
+        groups = generator._get_groups(config)
+
+        fd, tmp_file = tempfile.mkstemp()
+        f = open(tmp_file, 'w+')
+        formatter = generator._OptFormatter(output_file=f)
+        expected = '''[alpha]
+
+#
+# From namespace1
+#
+
+# foo option (string value)
+#foo = fred
+'''
+        generator._output_opts(formatter, 'alpha', groups.pop('alpha'))
+        f.close()
+        content = open(tmp_file).read()
+        self.assertEqual(expected, content)
 
 
 class GeneratorRaiseErrorTestCase(base.BaseTestCase):
@@ -835,4 +957,43 @@ class GeneratorRaiseErrorTestCase(base.BaseTestCase):
             self.assertRaises(cfg.RequiredOptError, generator.main, [])
 
 
+class ChangeDefaultsTestCase(base.BaseTestCase):
+
+    @mock.patch.object(generator, '_get_opt_default_updaters')
+    @mock.patch.object(generator, '_get_raw_opts_loaders')
+    def test_no_modifiers_registered(self, raw_opts_loaders, get_updaters):
+        orig_opt = cfg.StrOpt('foo', default='bar')
+        raw_opts_loaders.return_value = [
+            ('namespace', lambda: [(None, [orig_opt])]),
+        ]
+        get_updaters.return_value = []
+
+        opts = generator._list_opts(['namespace'])
+        # NOTE(dhellmann): Who designed this data structure?
+        the_opt = opts[0][1][0][1][0]
+
+        self.assertEqual('bar', the_opt.default)
+        self.assertIs(orig_opt, the_opt)
+
+    @mock.patch.object(generator, '_get_opt_default_updaters')
+    @mock.patch.object(generator, '_get_raw_opts_loaders')
+    def test_change_default(self, raw_opts_loaders, get_updaters):
+        orig_opt = cfg.StrOpt('foo', default='bar')
+        raw_opts_loaders.return_value = [
+            ('namespace', lambda: [(None, [orig_opt])]),
+        ]
+
+        def updater():
+            cfg.set_defaults([orig_opt], foo='blah')
+
+        get_updaters.return_value = [updater]
+
+        opts = generator._list_opts(['namespace'])
+        # NOTE(dhellmann): Who designed this data structure?
+        the_opt = opts[0][1][0][1][0]
+
+        self.assertEqual('blah', the_opt.default)
+        self.assertIs(orig_opt, the_opt)
+
+
 GeneratorTestCase.generate_scenarios()
